/*    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2013-
 *    Salvatore Ruggieri
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package dd.papers.tdp2014.dAnonymity;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.TreeSet;

import ags.utils.dataStructures.MaxHeap;
import ags.utils.dataStructures.trees.thirdGenKD.KdTree;
import ags.utils.dataStructures.trees.thirdGenKD.SquareEuclideanDistanceFunction;

import dd.util.ContingencyTable;
import dd.util.ContingencyTable.Measure;
import dd.util.DDTable;

import rst.table.Table;
import rst.table.TableSubset;

import rst.util.IOUtil;
import rst.util.Cronos;
import rst.util.data.Bitmap;
import rst.util.data.MDouble;
import rst.util.data.PairIntInt;
import rst.util.data.Triple;

/*
 * DMondrian: command line program for producing a bd(t)-protective dataset.
 */
public class DAnonymity {
	/** Input table for discrimination discovery (includes BitTable). */
	public DDTable tb;
	/** Input table in relational format. */
	public Table rtb;
	/** Compute t-closeness wrt SD measures. */
	boolean sdlike;
	/** Utility contingency tables for all PD itemsets. */
	ContingencyTable []cts;
	/** Non PD or Decision attributes. */
	ArrayList<Integer> nonPdDecAtts;
	/** PD or Decision attributes. */
	TreeSet<Integer> pdDecAtts;
	/** Output file name. */
	private String outputFile;
	/** Output. */
	PrintWriter writer;
	
	/** Constructor. */
	public DAnonymity(DDTable tb, Table rtb, boolean sdlike) {
		this.tb = tb;
		this.rtb = rtb;
		this.sdlike = sdlike;
	    cts = new ContingencyTable[tb.specs.length];
	    for(int i=0;i<tb.specs.length;++i)
	    	cts[i] = new ContingencyTable(tb.specs[i]);
	    nonPdDecAtts = new ArrayList<Integer>();
	    pdDecAtts = new TreeSet<Integer>();
	    for(int item : tb.pdItems)
	    	pdDecAtts.add( tb.dec.getAttribute(item) );
	    pdDecAtts.add( tb.dec.getAttribute(tb.decisionItem) );
	    int ncols = tb.getNoAtts();
		for(int i=0; i<ncols; ++i)
			if(!pdDecAtts.contains(i))
				nonPdDecAtts.add(i);
		tb.setTRanges(sdlike);
	}
		
	/** Set output of anonymization algorithm.  */
	private void setAnonymizeOutput(String output) throws IOException {
		writer = IOUtil.getWriter(tb.wDir+"/"+output);
		int ncols = tb.getNoAtts();
		/* Write header. */
		for(int i=0; i<ncols; ++i) {
			if(i>0) writer.print(",");
			writer.print(tb.dec.getAttName(i));
		}
		writer.println();
		this.outputFile = output;
	}

	/** Output sanitized subset. */
	public void output(TableSubset subset) {
		int ncols = tb.getNoAtts();
		PairIntInt[] range = subset.getRangeIdx(rtb);
		for(int pos : subset.rows) {
			for(int i=0; i<ncols; ++i) {
				if(i>0) writer.print(",");
				if( pdDecAtts.contains(i) )
					writer.print( rtb.columns.get(i).getString(pos) );
				else
//					writer.print( "[" + rtb.columns.get(i).getIdxValues(range[i].first) 
//							+ ":" + rtb.columns.get(i).getIdxValues(range[i].second) + "]");
					writer.print( "[" + range[i].first 
							+ ":" + range[i].second + "]");
			}
			writer.println();				
		}
	}
	
	/** Anonymization algorithm.  */
	public void anonymize(double tValue, boolean sabre, int k, Measure measure) throws Exception {
		// actual anonymization
		if(sabre) {
			if( k == Integer.MAX_VALUE )
				anonymizeDSabre(tValue);
			else
				anonymizeDSabreK(tValue, measure, true); // binary search
		}
		else
			anonymizeDMondrian(new TableSubset(rtb), tValue);
		writer.close();
	}

	/** dMondrian anonymization algorithm. */
	private void anonymizeDMondrian(TableSubset subset, double tValue) {
		double bestT = Double.MAX_VALUE; // best maximum variational distance so far
		int attBest = -1; // best attribute for cut
		int medianBest = -1; // median position of best attribute
		for(int i:nonPdDecAtts) { // for each non PD or Decision attribute, try a cut
			int median = subset.getMedian(rtb, i);
//			System.out.println("Att: " +i+" Median: "+median);
			if( median < 0 ) // only one distinct value
				continue;
			double maxT = -1; // maxT over all PD itemsets contingency tables
			for(ContingencyTable ct : cts) {
				TableSubset left = subset.split(rtb, i, median, true);
	//			System.out.println("Att: " + i+" Left: "+left.rows.size());
				tb.setCT(ct, left);
				double leftT = ct.tval(sdlike);
				if(leftT > tValue ) // not d-allowable
					continue; 
				TableSubset right = subset.split(rtb, i, median, false);
	//			System.out.println(i+" "+right.rows.size());
				tb.setCT(ct, right);
				double rightT = ct.tval(sdlike);
				if(rightT > tValue ) // not d-allowable 
					continue;
				double maxTct = leftT < rightT ? rightT : leftT; // maximum variational distance
				if(maxTct > maxT) // over all contingency tables
					maxT = maxTct;
			}
			if( maxT >= 0 && maxT < bestT) { // update best
				bestT = maxT;
				attBest = i;
				medianBest = median;
			}
		}
		if( attBest == -1 ) {
			// no best split - CSV output
			output(subset);
			return;
		}
		//System.out.println("Splitting att: " + rtb.columns.get(attBest).name + " on " + medianBest);
		anonymizeDMondrian(subset.split(rtb, attBest, medianBest, true), tValue);
		anonymizeDMondrian(subset.split(rtb, attBest, medianBest, false), tValue);
	}
	
	/** dSabre anonymization algorithm. */
	private void anonymizeDSabre(double tValue) {
		if(cts.length>1)
			throw new RuntimeException("dSabre not implemented for multiple PD items");
		ContingencyTable ct = cts[0]; 
		tb.setCT(ct);
		// buckets
		setBuckets(ct);
		// ECSize
		dichotomize(ct, tValue);
	}

	/** dSabre buckets. */
	KdTree<Integer> kdbucketPDneg;
	/** dSabre buckets. */
	KdTree<Integer> kdbucketPDpos;
	/** dSabre buckets. */
	KdTree<Integer> kdbucketPNDneg;
	/** dSabre buckets. */
	KdTree<Integer> kdbucketPNDpos;
	
	/** Initialize dSabre buckets. */
	private void setBuckets(ContingencyTable ct) {
		int k = nonPdDecAtts.size();
		kdbucketPDneg = new KdTree<Integer>(k);
		kdbucketPDpos = new KdTree<Integer>(k);
		kdbucketPNDneg = new KdTree<Integer>(k);
		kdbucketPNDpos = new KdTree<Integer>(k);
		int nrows = ct.n();
		Bitmap pdcover = ct.specs.coverPD;
		Bitmap negcover = tb.getCoverDecisionItem();
		for(int i=0; i<nrows; ++i)
			if( negcover.get(i) )
				if( pdcover.get(i) )
					kdbucketPDneg.addPoint(rtb.getEuclidian(i, nonPdDecAtts), i);
				else
					kdbucketPNDneg.addPoint(rtb.getEuclidian(i, nonPdDecAtts), i);
			else
				if( pdcover.get(i) )
					kdbucketPDpos.addPoint(rtb.getEuclidian(i, nonPdDecAtts), i);
				else
					kdbucketPNDpos.addPoint(rtb.getEuclidian(i, nonPdDecAtts), i);
		kdbucketPDneg.prepareDeletePoints();
		kdbucketPNDneg.prepareDeletePoints();
		kdbucketPDpos.prepareDeletePoints();
		kdbucketPNDpos.prepareDeletePoints();
		// debug checking
		if( ct.a != kdbucketPDneg.size() ||
				ct.b != kdbucketPDpos.size() ||
				ct.c != kdbucketPNDneg.size() ||
				ct.d != kdbucketPNDpos.size() )
			throw new RuntimeException("setBuckets inconsistency checked " + ct);
	}

	/** Take out from buckets elements as in a passed contingency table. */
	private void takeout(ContingencyTable ct) {
//		System.out.println("Picking up: " + ct + " from [" + kdbucketPDneg.size() + ", " + 
//				kdbucketPDpos.size() + ", " + kdbucketPNDneg.size() + ", " + kdbucketPNDpos.size() + "] ") ;
		TableSubset subset = new TableSubset();
		ArrayList<Triple<Double, Integer, KdTree<Integer>>> buckets = new ArrayList<Triple<Double, Integer, KdTree<Integer>>>();
		buckets.add(new Triple<Double, Integer, KdTree<Integer>>((double)kdbucketPDneg.size()/ct.a, ct.a, kdbucketPDneg)); // /ct.a + sort OK per credit e adult
		buckets.add(new Triple<Double, Integer, KdTree<Integer>>((double)kdbucketPDpos.size()/ct.b, ct.b, kdbucketPDpos));
		buckets.add(new Triple<Double, Integer, KdTree<Integer>>((double)kdbucketPNDneg.size()/ct.c, ct.c, kdbucketPNDneg));
		buckets.add(new Triple<Double, Integer, KdTree<Integer>>((double)kdbucketPNDpos.size()/ct.d, ct.d, kdbucketPNDpos));
		// senza sort OK per income
		Collections.sort(buckets); // search larger sets first
		for(Triple<Double, Integer, KdTree<Integer>> b : buckets)
			knn(b.second, b.third, subset);
		output(subset);
	}
	
	/** Nearest-neighboor search. */
	private void knn(int n, KdTree<Integer> kd, TableSubset subset) {
		final SquareEuclideanDistanceFunction df = new SquareEuclideanDistanceFunction();
		if(n==0) return;
		if(n > kd.size()) // debug check
			throw new RuntimeException("knn - too few elements: n = " + n + " from bn = " + kd.size());
		ArrayList<Integer> rows = subset.rows;
		if(rows.size()==0) {
			Integer r = kd.getRandomPoint();
			kd.deletePoint(r);
			rows.add( r );
			if(--n==0) return;
		}
		int x = rows.get(0);
		double []xarray = rtb.getEuclidian(x, nonPdDecAtts);
		MaxHeap<Integer> nearest = kd.findNearestNeighbors(xarray, n, df);
		int nn = nearest.size();
		for(int i=0; i<nn; ++i) {
			int row = nearest.getMax();
			nearest.removeMax();
			kd.deletePoint(row);
			rows.add(row);
		}
	}

	/** dSabre dichotomize. */
	private void dichotomize(ContingencyTable ct, double tValue) {
		int ctn = ct.n();
		if( ctn < 2) {
			takeout(ct);
			return;
		}
		// split
		ContingencyTable left = new ContingencyTable(ct.specs);
		ContingencyTable right = new ContingencyTable(ct.specs);
		
		left.a = (int)Math.floor(ct.a*0.5); 
		left.b = (int)Math.floor(ct.b*0.5); 
		left.c = (int)Math.ceil(ct.c*0.5); 
		left.d = (int)Math.ceil(ct.d*0.5); 
		int leftn = left.n();
		if(leftn == ctn) { // base case (0,0,1/0,1/0)
			if(left.c==1)
				left.c=0;
			else if(left.d==1)
				left.d=0;
			else throw new RuntimeException("dichotomize error");
		} else if(leftn==0) { // dual base case (1,1,0,0)
			if(ct.a==1)
				left.a=1;
			else if(ct.b==1)
				left.b=1;
			else throw new RuntimeException("dichotomize error");
		}
		double leftT = left.tval(sdlike);
		if(leftT > tValue ) { // left not t-close
			takeout(ct);
			return;
		}

		right.a = ct.a - left.a;
		right.b = ct.b - left.b;
		right.c = ct.c - left.c;
		right.d = ct.d - left.d;
		double rightT = right.tval(sdlike);
		if(rightT > tValue ) { // right not t-close
			takeout(ct);
			return;
		}
		dichotomize(left, tValue);
		dichotomize(right, tValue);
	}


	/** dSabreK anonymization algorithm. Binary search version. */
	private void anonymizeDSabreK(double alpha, Measure measure, boolean binarySearch) throws Exception {
		ContingencyTable ct = cts[0];
		tb.setCT(ct);
		
		// binary dichotomize
		final double EPSILON = 0.025;
		double tValue = tb.getBoundInverse(measure, alpha);
		int tmin = 0;
		int tmax = (int)Math.rint((tb.tValueMax-tValue)/EPSILON)+1;
		System.out.println("binary search t in [" + (tmin*EPSILON+tValue) + ", " + (tmax*EPSILON+tValue) + ")");
		System.out.println("alpha bound = " + alpha);
		MDouble minv = new MDouble();
		MDouble maxv = new MDouble();
		boolean lower = measure == Measure.RC || measure == Measure.EC;
		double bestalpha = lower ? Double.MIN_VALUE : Double.MAX_VALUE;
		while(tmin < tmax) {
			int midt = binarySearch ? (tmin+tmax)/2 : tmin;
//			System.out.println("tmin = " + tmin + " tmax = " + tmax + " midt = " + midt);
			double midtValue = midt*EPSILON+tValue;
//			System.out.println("trying t = " + midtValue);
			PrintWriter tmp = writer;
			String filetmp = outputFile;
			setAnonymizeOutput("tmp.arff");
			// buckets
			setBuckets(ct);
			// dichotomize
			dichotomize(ct, midtValue);
			writer.close();
			writer = tmp;
			outputFile = filetmp;
			statistics("tmp.arff", measure, minv, maxv, lower, alpha);
			double midtalpha = lower ? minv.value : maxv.value; 
//			System.out.println("for t = " + midtValue + " alpha = " + midtalpha);
			if( (lower && midtalpha >= alpha) || (!lower && midtalpha <= alpha)) {
				tmin = midt+1;
				bestalpha = midtalpha;
				IOUtil.rename(tb.wDir+"/tmp.arff", tb.wDir+"/best.arff");
			}
			else
				tmax = midt-1;
		}
		writer.close();
		if( (lower && bestalpha >= alpha) || (!lower && bestalpha <= alpha) ) {
			System.out.println("best t = " + ((tmin-1)*EPSILON+tValue));
			IOUtil.rename(tb.wDir+"/best.arff", tb.wDir+"/"+outputFile);
			IOUtil.deleteDir(tb.wDir+"/tmp.arff");
			IOUtil.deleteDir(tb.wDir+"/tmp.decode");
			IOUtil.deleteDir(tb.wDir+"/tmp.fimi");
		}
		else {
			System.out.println("no admissible t");
			IOUtil.deleteDir(tb.wDir+"/"+outputFile);
		}
	}

	/** Statistics on current dataset. Return elapsed time. 
	 * @param stop_if_higher 
	 * @param stop_if_lower */
	public double statistics(String file, Measure measure, MDouble minval, MDouble maxval, boolean lower, double alpha) throws Exception 
	{
	   Cronos c = new Cronos();
	   c.start();
  	   DDTable tb = new DDTable();
	   tb.setWorkingDir(this.tb.wDir);
	   tb.loadFromArff(file, true);
	   tb.setDiscMetaData(this.tb.pdItemString, this.tb.decItemString, this.tb.support);
	   tb.closedItemset = true;
	   tb.extractItemsets();
	   // process closed itemsets
	   tb.initScan();
	   ContingencyTable ct = null;
	   double maxv = Double.MIN_VALUE;
	   double minv = Double.MAX_VALUE;
   	   double m = 0;
	   while( (ct = tb.nextCT()) != null) {
		   m = ct.getMeasure(measure);
 		   maxv = Math.max(maxv, m);
 		   minv = Math.min(minv, m);
 		   if( (lower && minv < alpha) || (!lower && maxv > alpha))
 			   break;
   	   }
	   tb.endScan();
	   c.stop();
	   double time = c.elapsed();
	   minval.value = minv;
	   maxval.value = maxv;
	   return time;
	}

	public static void main(String[] args) throws Exception {
        if ( (args.length < 7) || (args.length > 9))  {
            System.out.println("Usage: java DMondrian (dMondrian|dSabre) [-sdlike] wdir input.arff PDitem DecItem t output.csv");
            System.out.println("Usage: java DMondrian dSabreK k measure wdir input.arff PDitem DecItem alpha output.csv");
            System.exit(0);
        }
        boolean sabre = args[0].equals("dSabre") || args[0].equals("dSabreK");
        int base = 1;
        int k = Integer.MAX_VALUE;
        Measure measure = Measure.RD;
        boolean sdlike = false;
        if( args[0].equals("dSabreK") ) {
        	k = Integer.parseInt(args[base++]);
        	measure = ContingencyTable.getMeasureCode(args[base++]);
        	sdlike = ContingencyTable.isSDlike(measure);
        } else
	        if( args[base].equals("-sdlike") ) {
	        	sdlike = true;
	        	++base;
	        }
        
 		Cronos t = new Cronos();
 		t.start();
 		
 		/** Table for DD. */
		DDTable tb = new DDTable();
		tb.setWorkingDir(args[base]);
		tb.loadFromArff(args[base+1]);
		tb.setDiscMetaData(args[base+2], args[base+3], k);
	
 		/** Relational format able for output. */
		Table rtb = Table.loadFromArff(args[base]+"/"+args[base+1]);
		
		DAnonymity dm = new DAnonymity(tb, rtb, sdlike);
 		double tValue = Double.parseDouble(args[base+4]);
 		if(!args[0].equals("dSabreK") && (tValue < tb.tValueMin || tValue > tb.tValueMax)) {
 			System.out.println("t value ranges in ["+tb.tValueMin+", "+tb.tValueMax+"].");
 			System.out.println("[Warning] t value " + tValue + " out of range.");
 		}
 		dm.setAnonymizeOutput(args[base+5]);
 		dm.anonymize(tValue, sabre, k, measure);
 		
 		t.stop();
 		System.out.println("Elapsed time: " + t.elapsed());

	}
}
