/*    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2008-
 *    Salvatore Ruggieri
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package rst.pp.fimi;

import java.io.*;
import java.util.*;

import rst.util.IOUtil;
import rst.util.data.MInteger;

/** Decode integers item(set)s from an encoding file. */
public class Decode 
{
	/** Map a string item into its code. */
	TreeMap<String, Integer> deco;
	/** String of an item code. */
	String[] code;
	/** Support of item code. */
	int[] supp;
	/** Number of attributes. */
	int noattribute;
	/** Name of attributes. */
	String[] attName;	
	/** Attribute number of item code. */
	int[] attNo;
	/** Number of items per attribute. */
	int[] attributeNitems;
	/** Mark removed items. */
	boolean[] removed;
	/** Number of an attribute name. */
	TreeMap<String, Integer> attcode;
	
	/** Constructor. */
	public Decode(String encodingFileName) 
	throws IOException
	{
		BufferedReader in = IOUtil.getReader(encodingFileName);
		// count number of codes
		int maxv = 0;
		int minv = Integer.MAX_VALUE;
		String[] linev;
		while ( (linev = IOUtil.readTokens(in, "\t")) != null) {
			int n = Integer.parseInt(linev[0]);
			if( n > maxv )
				maxv = n;
			if( n < minv )
				minv = n;
		}
		in.close();
		// allocate arrays
		code = new String[++maxv];
		supp = new int[maxv];
		removed = new boolean[maxv];
		// store codes and decodes
		in = IOUtil.getReader(encodingFileName);
		deco = new TreeMap<String, Integer>();
		while ( (linev = IOUtil.readTokens(in, "\t")) != null) {
			int n = Integer.parseInt(linev[0]);
			code[n] = linev[1];
			supp[n] = Integer.parseInt(linev[2]);
			deco.put(code[n], new Integer(n));
		}
		in.close();
		// computes attributes
		attNo = new int[maxv];
		noattribute = 0;
		attcode = new TreeMap<String, Integer>();
		TreeMap<Integer, MInteger> attnoitems = new TreeMap<Integer, MInteger>();
		for(int i=minv; i<maxv; ++i) {
			String attributeName = attributeName(code[i]);
			Integer attPos =  attcode.get(attributeName);
			if( attPos == null ) {
				attcode.put(attributeName, noattribute);
				attnoitems.put(noattribute, new MInteger(1));
				attNo[i] = noattribute;
				++noattribute;
			} else {
				attNo[i] = attPos;
				attnoitems.get(attPos).value++;
			}
		}
		attName = new String[noattribute];
		for( Map.Entry<String,Integer> e : attcode.entrySet() )
			attName[ e.getValue() ] = e.getKey();		
		attributeNitems = new int[noattribute];
		for( Map.Entry<Integer,MInteger> e : attnoitems.entrySet() )
			attributeNitems[ e.getKey() ] = e.getValue().value;	
	}
	/** Number of attributes. */
	public int getNoAtts()
	{
		return noattribute;
	}
	/** Number of items for a given attribute. */
	public int getAttributeNItems(int attno) {
		return attributeNitems[attno];
	}
	/** Attribute name. */
	public String getAttName(int attNo)
	{
		return attName[attNo];
	}
	/** Attribute number. */
	public int getAttNo(String attributeName) {
		Integer attPos =  attcode.get(attributeName);
		if(attPos==null)
			return -1;
		return attPos;
	}
	/** Attribute of an encoded item. */
	public int getAttribute(int item)
	{
		return attNo[item];
	}
	/** Support of an encoded item. */
	public int support(int item) 
	{
		return supp[item];		
	}
	/** Number of items encoded. */
	public int numberOfCodes() {
		return code.length;
	}
	/** Decode an encoded item. */
	public String decode(int item) 
	{
		return code[item];		
	}
	/** Decode an itemset, with items separated by sep. */
	public String decode(String items, String sep) 
	{
		StringTokenizer tk = new StringTokenizer(items, sep);
		String res = "";
		boolean flag = true;
		while(tk.hasMoreTokens()) {
			String token = tk.nextToken();
			try {
				int n = Integer.parseInt(token);
				if( n >= 0 && n < code.length )
					res += " " + code[n];
				else
					res += " " + token;
			} catch(NumberFormatException e ) { 
				res += (flag ? "" : " ") + token;
				flag = false;
			}
		}
		return res;		
	}
	/** Decode an itemset, with items separated by space. */
	public String decode(String items) 
	{
		return decode(items, " ");
	}
	/** Encode an item. */
	public int encode(String item) 
	{
		Integer n = deco.get(item);
		if( n==null )
			throw new RuntimeException("encoding " + item);
		return n.intValue();
	}
	/** Encode an itemset. */
	public String encodeAll(String items) 
	{
		if( items.equals(""))
			return "";
		StringTokenizer tk = new StringTokenizer(items, " ");
		String res = "";
		boolean flag = true;
		while(tk.hasMoreTokens()) {
			String token = tk.nextToken();
			res += (flag ? "" : " ") + encode(token);
			flag = false;
		}
		return res;		
	}
	/** Extract attribute name from an item. */
	private static String attributeName(String item)
	{
		return item.substring(0, item.indexOf("="));
	}
	/** Check whether an item is marked as removed. */
	public boolean isRemoved(int item)
	{
		return removed[item];
	}
	/** Mark an item as removed or not. */
	public void setRemoved(int item, boolean value)
	{
		removed[item] = value;
	}
	/** Mark an item as removed or not. */
	public void setRemoved(String item, boolean value)
	{
		removed[ deco.get(item).intValue() ] = value;
	}
	/** Check whether is there any removed item. */
	public boolean isAnyRemoved()
	{
		for(int i=0; i<code.length; ++i)
			if( removed[i] )
				return true;
		return false;
	}
	/** Parse a FIMI file, removing items marked as removed. */
	public void removeItems(String inFimiFile, String outFimiFile) throws IOException
	{
		BufferedReader in = IOUtil.getReader(inFimiFile);
		PrintWriter out = IOUtil.getWriter(outFimiFile);
		String line;
		while( (line=in.readLine()) != null ) {
			StringTokenizer tk = new StringTokenizer(line, " ");
			boolean flag = true;
			String outline = "";
			while(tk.hasMoreTokens()) {
				String token = tk.nextToken();
				int item = Integer.parseInt(token);
				if( removed[item] )
					continue;
				if( flag ) {
					outline = token;
					flag = false;
				} else
					outline += " " + token;					
			}
			if( !flag )
				out.println(outline);
		}
		in.close();
		out.close();
	}
	
	/** Test command line. */
	public static void main(String[] args) {
	    if (args.length != 3) {
	        System.out.println("Usage: java Decode input.decode input output");
	            System.exit(0);
	        }
	        try {
	        	Decode d = new Decode(args[0]);
	            BufferedReader in = IOUtil.getReader(args[1]);
	            PrintWriter out = IOUtil.getWriter(args[2]);
	            String line;
	     	    while( (line = in.readLine() ) != null ) 
	     	    	out.println(d.decode(line));
	            in.close();
	            out.close();
	        } catch (IOException e) {
	            System.out.println(e.toString());
	       }
	 }
}
