/*    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2013-
 *    Salvatore Ruggieri
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package rst.table;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;

import rst.pp.fimi.CSV2FIMI;
import rst.pp.fimi.Decode;
import rst.util.*;
import rst.util.data.PairIntInt;

/** Relational table, with column stored using indexes to values. */
public class Table {
	/** Table name. */
	public String name;
	/** All columns. */
	public ArrayList<Column> columns;
	/** Only columns of Doubles. */
	public ArrayList<DoubleColumn> realColumns;
	/** Only columns of Strings. */
	public ArrayList<StringColumn> enumeratedColumns;
	
	/** Constructor. */
	public Table(String name) {
		this.name = name;
		columns = new ArrayList<Column>();
		realColumns = new ArrayList<DoubleColumn>();
		enumeratedColumns = new ArrayList<StringColumn>();
	}
	
	/** Number of rows. */
	public int size() {
		if( columns.size() == 0 )
			return -1;
		return columns.get(0).index.length;
	}

	/** Add a column of Double. */
	public void addRealColumn(DoubleColumn doubleCol) {
		realColumns.add(doubleCol);
		columns.add(doubleCol);
	}

	/** Add a column of Strings. */
	public void addEnumeratedColumn(StringColumn enumCol) {
		enumeratedColumns.add(enumCol);
		columns.add(enumCol);
	}
	
	/** Return position of a column. Implementation is inefficient. */
	public int getAttPos(String colName) {
		for(int i=0; i<columns.size(); ++i) 
			if( columns.get(i).name.equals(colName) )
				return i;
		return -1;
	}

	/** Return a column of type DoubleColumn. Implementation is inefficient. */
	public DoubleColumn getAttDouble(String colName) {
		for(int i=0; i<realColumns.size(); ++i)
			if( realColumns.get(i).name.equals(colName) )
				return realColumns.get(i);
		return null;
	}

	/** Return a column of type StringColumn. Implementation is inefficient. */
	public StringColumn getAttString(String colName) {
		for(int i=0; i<realColumns.size(); ++i)
			if( enumeratedColumns.get(i).name.equals(colName) )
				return enumeratedColumns.get(i);
		return null;
	}

	/** Return ranges of indexes for all columns. */
	public PairIntInt[] getRanges() {
		int cols = columns.size();
		PairIntInt[] ranges = new PairIntInt[cols];
		for(int i=0; i<cols; ++i) {
			ranges[i] = new PairIntInt(0, columns.get(i).getLastIndex());
		}
		return ranges;
	}
	
	/** Select rowid From table Where C1 in [min,max], C2 in ... */
	public TableSubset select(PairIntInt[] ranges) {
		TableSubset sub = new TableSubset();
		int rows = size();
		int cols = columns.size();
		for(int i=0;i<rows;++i) {
			boolean include = true;
			for(int j=0; j<cols; ++j) {
				int ij = columns.get(j).index[i];
				if(ij < ranges[j].first || ij > ranges[j].second) {
					include = false;
					break;
				}
			}
			if(include)
				sub.rows.add(i);
		}
		return sub;
	}
	
	/** Select count(*) From table Where C1 in [min,max], C2 in ... */
	public int selectCount(PairIntInt[] ranges) {
		int sub = 0;
		int rows = size();
		int cols = columns.size();
		for(int i=0;i<rows;++i) {
			boolean include = true;
			for(int j=0; j<cols; ++j) {
				int ij = columns.get(j).index[i];
				if(ij < ranges[j].first || ij > ranges[j].second) {
					include = false;
					break;
				}
			}
			if(include)
				++sub;
		}
		return sub;
	}

	// to be documented
	private final java.util.Random ra = new java.util.Random(System.currentTimeMillis());
	private final java.util.Random r = new java.util.Random(System.currentTimeMillis());
	// to be documented
	public PairIntInt[] randomRanges(int nAttsMin, int nAttsMax, int mustBeAtt) {
		PairIntInt[] res = getRanges();
		int cols = columns.size();
		int toselect = cols-1;
		int nAtts = nAttsMin;
		if(nAttsMax != nAttsMin)
			nAtts += ra.nextInt(nAttsMax-nAttsMin+1);
		for(int i=0; i<cols; ++i) {
			if( i != mustBeAtt ) {
				int rn = r.nextInt(toselect);
				--toselect;
				if(rn>=nAtts)
					continue;
				--nAtts;
			}
			Column col = columns.get(i);
			int coln = col.getNDistinct();
			int len = r.nextInt(coln);
			int start = r.nextInt(coln-len);
			res[i].first = start;
			res[i].second = start + len;
		}
		if(nAtts!=0)
			throw new RuntimeException("randomRanges");
		return res;
	}

	/** Decode an item att=value -- seemns unused
	public PairIntInt getAttIdx(String attValue) {
		int mid = attValue.indexOf("=");
		int att = getAttPos(attValue.substring(0, mid));
		int idx = columns.get(att).getIdx(attValue.substring(mid+1));
		return new PairIntInt(att, idx);
	}*/

	/** Load table from Arff. */
	public static Table loadFromArff(String fileName) throws IOException {
		BufferedReader arff = IOUtil.getReader(fileName);
		// count number of tuples
		while( !arff.readLine().equals("@data") ) ;
		int size = 0;
		String []items;
		while( (items = IOUtil.readTokens(arff)) != null) {
			if( items.length == 0)
				continue;
			++size;
		}
		arff.close();
		// read table name
		arff = IOUtil.getReader(fileName);
		String line;
		while( !(line = arff.readLine()).startsWith("@relation") ) ;
		String[] row = line.split(" ");
		Table tb = new Table( row.length > 1 ? row[1] : "myTable" );
		// read column names and types
		while( (line = arff.readLine()) != null ) {
			if( line.equals("@data") )
				break;
			if( !line.startsWith("@attribute") )
				continue;
			int skip = 1+"@attribute".length();
			int space = line.indexOf(" ", skip);
			String colName = line.substring(skip, space);
			skip = space + 1;
			space = line.indexOf(" ", skip);
			if(space<0) space = line.length();
			String colType = line.substring(skip, space).trim();
//			System.out.println("attribute " + colName + " " + colType);
			if( colType.equals("real") || colType.equals("numeric") || colType.equals("integer") ) {
				DoubleColumn doubleCol = new DoubleColumn(colName, size);
				tb.addRealColumn(doubleCol);
			} else if( colType.startsWith("{") ) {
				space = line.lastIndexOf("}");
				String enums = line.substring(skip+1,space).trim();
				StringColumn enumCol = new StringColumn(colName, size);
				tb.addEnumeratedColumn(enumCol);
				// parse order of enumeration
				items = enums.split(",");
				for( String item : items )
					enumCol.addValue(item.trim());
				
			} else throw new RuntimeException("Reading arff file: unhandled column type " + colType);
		}
		int ncols = tb.columns.size();
		int pos = 0;
		while( (items = IOUtil.readTokens(arff)) != null) {
			if( items.length == 0)
				continue;
//			System.out.println(items.length);
			for(int i=0; i<ncols; ++i) {
				String item = items[i].trim();
				tb.columns.get(i).setString(pos, item);
			}
			++pos;
		}
		// re-order index values for Double columns
		for(DoubleColumn doubleCol : tb.realColumns )
			doubleCol.sortCodes();
		return tb;
	}
	
	/** Output table in CSV format to standard output. */
	public void toCSV() throws IOException {
		toCSV(IOUtil.getWriter(""));
	}
	
	/** Output table in CSV format to output. */
	public void toCSV(PrintWriter out) {
		int ncols = columns.size();
		int size = this.size();
		for(int i=0; i<size; ++i) {
			for(int j=0;j<ncols; ++j) {
				if( j > 0)
					out.print(",");
				out.print( columns.get(j).getString(i));
			}
			out.println();
		}
		
	}
	
	/** Output table in CSV format to output. */
	public void toCSV(TableSubset subset, PrintWriter out, String sep) {
		int ncols = columns.size();
		int size = subset.size();
		for(int i=0; i<size; ++i) {
			for(int j=0;j<ncols; ++j) {
				if( j > 0)
					out.print(sep);
				out.print(columns.get(j).getString(subset.rows.get(i)));
			}
			out.println();
		}
		
	}
	
	/** Output table first columns in CSV format to output. */
	public void toCSVprefix(TableSubset tuples, PrintWriter out, int ncols, String sep) {
		int size = tuples.size();
		for(int i=0; i<size; ++i) {
			for(int j=0;j<ncols; ++j) {
				if( j > 0)
					out.print(sep);
				out.print(columns.get(j).getString(tuples.rows.get(i)));
			}
			out.println();
		}
	}

	/** Manhattan distance. */
	public double distanceManhattan(int pos1, int pos2, ArrayList<Integer> atts) {
		double d = 0;
		for(int a : atts) {
			double ad = columns.get(a).distanceManhattan(pos1, pos2);
			d += ad;
		}
		return d;
	}
	
	/** Get multidimensional point in normalized Euclidian space. */
	public double[] getEuclidian(int row, ArrayList<Integer> atts) {
		int sz = atts.size();
		double []res = new double[sz];
		for(int i=0; i<sz; ++i) {
			Column col = columns.get(atts.get(i));
			res[i] = col.index[row]/(double)(col.getNDistinct()-1);
		}
		return res;
	}

	/** Normalized quadratic Euclidian distance. */
	public double distanceNormalizedEuclidian(int pos1, int pos2, ArrayList<Integer> atts) {
		double d = 0;
		for(int a : atts) {
			double ad = columns.get(a).distanceNormalizedEuclidian(pos1, pos2);
			d += ad;
		}
		return d;
	}
	
	/** Quadratic Euclidian distance. */
	public double distanceEuclidian(int pos1, int pos2, ArrayList<Integer> atts) {
		double d = 0;
		for(int a : atts) {
			double ad = columns.get(a).distanceEuclidian(pos1, pos2);
			d += ad;
		}
		return d;
	}
	
	/** Test method. */
	public static void main(String []argv) throws IOException {
		String tableName = "income14";
		Cronos c = new Cronos();
		c.start();
		Table tb = Table.loadFromArff(tableName+".arff");
		System.out.println("Table size: " + tb.size() );
		System.out.println("Table size by select: " + tb.select(tb.getRanges()).size() );
		//tb.toCSV(IOUtil.getWriter(tableName+".csv"));
		c.stop();
		System.out.println("Table loaded - time (s): " + c.elapsed());
		c.start();
		String cmd = tableName+".arff " + tableName + ".fimi " + tableName + ".decode";
		CSV2FIMI.main(cmd.split(" "));
		c.stop();
		System.out.println("CSV2FIMI done - time (s): " + c.elapsed());
		c.start();
		Decode dec = new Decode(tableName + ".decode");
		BitTable bt = new BitTable(dec.numberOfCodes(), tb.size());
		bt.loadFromFimi(tableName+".fimi");
		c.stop();
		System.out.println("BitTable loaded - time (s): " + c.elapsed());
		c.start();
		int age = dec.encode("race=White");
		int worker = dec.encode("class_of_worker=Private");
		int income = dec.encode("class=-_50000.");
		System.out.println("Support count: " + bt.supp(new int[] {age, worker, income}) );
		c.stop();
		System.out.println("Support count - time (s): " + c.elapsed());
		
	}

}
